"""
Module with classes for tag selection based on CSS selectors.

Contains implementation of basic CSS pseudo-classes like
`:only-child`, `:empty`, `:nth-child()`.

They can be used in combination with other `SoupSelector` objects
to create more complex search procedures.

Classes
-------
- OnlyChild
- Empty
- FirstChild
- LastChild
- NthChild
- NthLastChild
- FirstOfType
- LastOfType
- NthOfType
- NthLastOfType
- OnlyOfType
- CSS - wrapper for simple search with CSS selectors
"""

from itertools import islice
from typing import Optional

try:
    import soupsieve as sv
except ImportError as e:
    raise ImportError(
        "`soupsavvy.css` requires `soupsieve` package to be installed."
    ) from e
from bs4 import Tag

import soupsavvy.exceptions as exc
from soupsavvy.base import SelectableCSS, SoupSelector
from soupsavvy.utils.selector_utils import TagIterator


class CSSSoupSelector(SoupSelector, SelectableCSS):
    """
    Base class for selectors based on css selectors.
    Classes that base their selection on CSS selectors should inherit from this class
    and implement the `selector` property.

    By default, the `find` methods are implemented using the `soupsieve` library
    to match the selector.

    CSSSoupSelector objects inherit from `SoupSelector` and can be easily used
    in combination with other `SoupSelector` objects.
    """

    SELECTOR: str

    def __init__(self) -> None:
        selector = self.__class__.SELECTOR.format(*self._formats)

        try:
            self._compiled = sv.compile(selector)
        except sv.SelectorSyntaxError:
            raise exc.InvalidCSSSelector(
                "CSS selector constructed from provided parameters "
                f"is not valid: {selector}"
            )
        self._selector = selector

    @property
    def _formats(self) -> list[str]:
        """
        Returns list of arguments to be used to format css selector string.
        """
        return []

    @property
    def css(self) -> str:
        return self._selector

    def find_all(
        self,
        tag: Tag,
        recursive: bool = True,
        limit: Optional[int] = None,
    ) -> list[Tag]:

        iterator = TagIterator(tag, recursive=recursive)
        return list(islice(filter(self._compiled.match, iterator), limit))

    def __eq__(self, other: object) -> bool:
        # we only care if selector is equal with current implementation
        return isinstance(other, CSSSoupSelector) and self.css == other.css


class OnlyChild(CSSSoupSelector):
    """
    Selector for finding elements, that do not have any siblings.
    Counterpart of the CSS selector `:only-child` pseudo-class.

    Example
    --------
    >>> <div class="widget"> ❌
    ...    <div class="menu">Hello World</div> ✔️
    ... </div>
    ... <div class="widget"> ❌
    ...    <div class="menu">Hello World</div> ❌
    ...    <div class="menu">Hello World 2 </div> ❌
    ... </div>

    Notes
    --------
    For more information on the :only-child pseudo-class, see:

    https://developer.mozilla.org/en-US/docs/Web/CSS/:only-child
    """

    SELECTOR = ":only-child"


class Empty(CSSSoupSelector):
    """
    Selector for finding empty elements, i.e., that have no children.
    Counterpart of the CSS selector `:empty` pseudo-class.

    Example
    --------
    >>> <div class="widget"> ❌
    ...    <div class="menu"></div> ✔️
    ... </div>

    Any text node is considered as a child and makes the element non-empty.

    Example
    --------
    >>> <div class="widget">Hello World</div> ❌
    ... <div class="widget"> </div> ❌

    These elements are not empty and do not match the selector.

    Notes
    --------
    For more information on the :empty selector, see:

    https://developer.mozilla.org/en-US/docs/Web/CSS/:empty
    """

    SELECTOR = ":empty"


class FirstChild(CSSSoupSelector):
    """
    Selector for finding elements, that are the first child of their parent.
    Counterpart of the CSS selector `:first-child` pseudo-class.

    Example
    --------
    >>> <div class="widget"> ✔️
    ...    <div class="menu">Hello World</div> ✔️
    ... </div>
    ... <div class="widget"> ❌
    ...    <div class="menu">Hello World</div> ✔️
    ...    <div class="menu">Hello World 2 </div> ❌
    ... </div>

    Notes
    --------
    `FirstChild` object is essentially the same as NthChild("1").

    For more information on the :first-child selector, see:

    https://developer.mozilla.org/en-US/docs/Web/CSS/:first-child
    """

    SELECTOR = ":first-child"


class LastChild(CSSSoupSelector):
    """
    Selector for finding elements, that are the last child of their parent.
    Counterpart of the CSS selector `:last-child` pseudo-class.

    Example
    --------
    >>> <div class="widget"> ❌
    ...    <div class="menu">Hello World</div> ✔️
    ... </div>
    ... <div class="widget"> ✔️
    ...    <div class="menu">Hello World</div> ❌
    ...    <div class="menu">Hello World 2 </div> ✔️
    ... </div>

    Notes
    --------
    LastChild object is essentially the same as NthLastChild("1").
    Element that is the first and only child is matched as well.

    For more information on the :last-child selector, see:

    https://developer.mozilla.org/en-US/docs/Web/CSS/:last-child
    """

    SELECTOR = ":last-child"


class NthBaseSelector(CSSSoupSelector):
    """Base class for selectors based on nth formula."""

    def __init__(self, nth: str) -> None:
        """
        Initialize the selector with the nth formula.

        Parameters
        ----------
        nth : str, positional
            Nth formula used to select the elements.
        """
        self._nth = nth
        super().__init__()

    @property
    def _formats(self) -> list[str]:
        return [self._nth]


class NthChild(NthBaseSelector):
    """
    Selector for finding elements, that are the nth child of their parent.
    Counterpart of the CSS selector `:nth-child(n)`.

    Example
    --------
    >>> NthChild("2").selector
    :nth-child(2)
    >>> NthChild("2n+1").selector
    :nth-child(2n+1)
    >>> NthChild("odd").selector
    :nth-child(odd)

    Notes
    --------
    For more information on the formula, see:

    https://developer.mozilla.org/en-US/docs/Web/CSS/:nth-child
    """

    SELECTOR = ":nth-child({})"


class NthLastChild(NthBaseSelector):
    """
    Selector for finding elements, that are the nth last child of their parent.
    Counterpart of the CSS selector `:nth-last-child(n)`.

    Example
    --------
    >>> NthLastChild("2").selector
    :nth-last-child(2)
    >>> NthLastChild("2n+1").selector
    :nth-last-child(2n+1)
    >>> NthLastChild("odd").selector
    :nth-last-child(odd)

    Notes
    --------
    For more information on the formula, see:

    https://developer.mozilla.org/en-US/docs/Web/CSS/:nth-last-child
    """

    SELECTOR = ":nth-last-child({})"


class FirstOfType(CSSSoupSelector):
    """
    Selector for finding elements, that are the first of their type in their parent.
    Counterpart of the CSS selector `:first-of-type` pseudo-class.

    Example
    --------
    >>> <div class="widget"> ✔️
    ...    <div class="menu">Hello World</div> ✔️
    ... </div>
    ... <div class="widget"> ❌
    ...    <div class="menu">Hello World</div> ✔️
    ...    <span class="menu">Hello World 2 </span> ✔️
    ...    <div class="menu">Hello World 3 </div> ❌
    ... </div>

    Notes
    --------
    For this selector the first tag of any type is selected, which in
    case of finding single tag is equivalent to `FirstChild` results.

    For more information on the formula, see:

    https://developer.mozilla.org/en-US/docs/Web/CSS/:first-of-type
    """

    SELECTOR = ":first-of-type"


class LastOfType(CSSSoupSelector):
    """
    Selector for finding elements, that are the last of their type in their parent.
    Counterpart of the CSS selector `:last-of-type` pseudo-class.

    Example
    --------
    >>> <div class="widget"> ❌
    ...    <div class="menu">Hello World</div> ✔️
    ... </div>
    ... <div class="widget"> ✔️
    ...    <div class="menu">Hello World</div> ❌
    ...    <span class="menu">Hello World 2 </span> ✔️
    ...    <div class="menu">Hello World 3 </div> ✔️
    ... </div>

    Notes
    --------
    For this selector the last tag of any type is selected, which in
    case of finding single tag is the equivalent to `LastChild` results.

    For more information on the formula, see:

    https://developer.mozilla.org/en-US/docs/Web/CSS/:last-of-type
    """

    SELECTOR = ":last-of-type"


class NthOfType(NthBaseSelector):
    """
    Selector for finding elements, that are the nth of their type in their parent.
    Counterpart of the CSS selector `:nth-of-type(n)`.

    Example
    --------
    >>> NthOfType("2").selector
    :nth-of-type(2)
    >>> NthOfType("2n+1").selector
    :nth-of-type(2n+1)
    >>> NthOfType("even").selector
    :nth-of-type(even)

    Notes
    --------
    For more information on the formula, see:

    https://developer.mozilla.org/en-US/docs/Web/CSS/:nth-of-type
    """

    SELECTOR = ":nth-of-type({})"


class NthLastOfType(NthBaseSelector):
    """
    Selector for finding elements, that are the nth last of their type in their parent.
    Counterpart of the CSS selector `:nth-last-of-type(n)`.

    Example
    --------
    >>> NthLastOfType("2").selector
    :nth-last-of-type(2)
    >>> NthLastOfType("2n+1").selector
    :nth-last-of-type(2n+1)
    >>> NthLastOfType("even").selector
    :nth-last-of-type(even)

    Notes
    --------
    For more information on the formula, see:

    https://developer.mozilla.org/en-US/docs/Web/CSS/:nth-last-of-type
    """

    SELECTOR = ":nth-last-of-type({})"


class OnlyOfType(CSSSoupSelector):
    """
    Selector for finding elements, that don't have siblings of the same type.
    Counterpart of the CSS selector `:only-of-type` pseudo-class.

    Example
    --------
    >>> <div class="widget"> ❌
    ...    <div class="menu">Hello World</div> ✔️
    ... </div>
    ... <div class="widget"> ❌
    ...    <div class="menu">Hello World</div> ✔️
    ...    <span class="menu">Hello World 2 </span> ❌
    ...    <span class="menu">Hello World 3 </span> ❌
    ... </div>
    ... <a class="widget"></a> ✔️

    Notes
    --------
    For more information on the :only-of-type selector, see:

    https://developer.mozilla.org/en-US/docs/Web/CSS/:only-of-type
    """

    SELECTOR = ":only-of-type"


class CSS(CSSSoupSelector):
    """
    Selector for finding elements based on any provided CSS selector.
    `soupsieve` adapter, that allows any supported css selector
    to be used with other `soupsavvy` components.

    Example
    --------
    >>> CSS("div.menu")

    Would match:

    Example
    --------
    >>> <div class="widget"> ❌
    ...    <div class="menu">Hello World</div> ✔️
    ... </div>
    ... <div class="menu_main"> ❌
    ...    <a class="menu">Hello World</a> ❌
    ... </div>
    ... <div class="menu"></div> ✔️

    Notes
    --------
    `CSS` component extends `bs4.Tag.select` implementation
    by adding non recursive search option.
    """

    SELECTOR = "{}"

    def __init__(self, css: str) -> None:
        """
        Initializes the selector with the provided css selector.

        Parameters
        ----------
        css : str
            CSS selector to be used for selecting elements.
        """
        self._css = css
        super().__init__()

    @property
    def _formats(self) -> list[str]:
        return [self._css]



"""
Module provides `nth-of-type` selector implementations for `SoupSelector`.
It allows you to search for the nth occurrence of an element,
similar to how the CSS `nth-of-type` pseudo-class works.
However, instead of being limited to css selectors, it works with any `SoupSelector` instance.

Classes
-------
- `NthOfSelector` - Selects nth element matching given selector
- `NthLastOfSelector` - Selects nth last element matching given selector
- `OnlyOfSelector` - Selects only element matching given selector
"""

from typing import Optional

from bs4 import Tag

from soupsavvy.base import SoupSelector, check_selector
from soupsavvy.selectors.nth.nth_utils import parse_nth
from soupsavvy.utils.selector_utils import TagIterator, TagResultSet


class BaseNthOfSelector(SoupSelector):
    """
    Base class for nth-of-selector and nth-last-of-selector
    that implements general logic for finding matching elements.
    """

    # slice for modification of list of elements matching selector
    _slice: slice

    def __init__(self, selector: SoupSelector, nth: str) -> None:
        """
        Initializes nth selector instance.

        Parameters
        ----------
        selector : SoupSelector
            Any `SoupSelector` instance used to match elements.
        nth : str
            CSS nth selector string. Accepts all valid css nth formulas.

        Raises
        ------
        NotSoupSelectorException
            If selector is not an instance of `SoupSelector`.
        """
        self._selector = check_selector(selector)
        self.nth_selector = parse_nth(nth)

    @property
    def selector(self) -> SoupSelector:
        """
        Returns selector instance used for matching elements in this nth selector.

        Returns
        -------
        SoupSelector
            Selector used in this nth selector.
        """
        return self._selector

    def find_all(
        self,
        tag: Tag,
        recursive: bool = True,
        limit: Optional[int] = None,
    ) -> list[Tag]:
        # if recursive is False, check only children of tag itself
        tag_iterator = (
            TagIterator(tag, recursive=recursive, include_self=True)
            if recursive
            else iter([tag])
        )

        matches = []

        for tag_ in tag_iterator:
            matching = self.selector.find_all(tag=tag_, recursive=False)[self._slice]
            matches += [
                matching[index - 1]
                for index in self.nth_selector.generate(len(matching))
            ]

        # keep order of tags and limit
        results = TagResultSet(
            list(TagIterator(tag, recursive=recursive))
        ) & TagResultSet(matches)

        return results.fetch(limit)

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False

        return (
            self.selector == other.selector and self.nth_selector == other.nth_selector
        )

    def __repr__(self):
        cls = self.__class__.__name__
        return f"{cls}(selector={self.selector}, nth={self.nth_selector})"

    def __str__(self):
        return repr(self)


class NthOfSelector(BaseNthOfSelector):
    """
    Selector for finding nth-of elements in the soup among elements that match
    provided `SoupSelector` instance.

    Example
    -------
    >>> selector = NthOfSelector(ClassSelector("item"), "2n+1")

    matches all odd elements with class "item".

    Example
    -------
    >>> <div class="item">1</div> ✔️
    ... <div id="item"></div> ❌
    ... <div class="item">2</div> ❌
    ... <div class="item">3</div> ✔️
    ... <div class="widget"></div> ❌
    ... <div class="item">4</div> ❌

    Notes
    -----
    For more information about standard :nth-of-type pseudo-class, visit:

    https://developer.mozilla.org/en-US/docs/Web/CSS/:nth-of-type
    """

    # keep initial order of matching elements
    _slice = slice(None)


class NthLastOfSelector(BaseNthOfSelector):
    """
    Selector for finding nth-last-of elements in the soup among elements that match
    provided `SoupSelector` instance.

    Example
    -------
    >>> selector = NthLastOfSelector(ClassSelector("item"), "2n+1")

    matches all odd elements with class "item" starting from the last element.

    Example
    -------
    >>> <div class="item">1</div> ❌
    ... <div id="item"></div> ❌
    ... <div class="item">2</div> ✔️
    ... <div class="item">3</div> ❌
    ... <div class="widget"></div> ❌
    ... <div class="item">4</div> ✔️

    Notes
    -----
    For more information about standard :nth-of-type pseudo-class, visit:

    https://developer.mozilla.org/en-US/docs/Web/CSS/:nth-last-of-type
    """

    # reverse order of matching elements
    _slice = slice(None, None, -1)


class OnlyOfSelector(SoupSelector):
    """
    Selector for finding the only element,
    that matches provided `SoupSelector` instance among its siblings.

    Example
    -------
    >>> selector = OnlyOfSelector(ClassSelector("item"))

    matches all elements with class "item" that are the only child of their parent
    that matches the selector.

    Example
    -------
    >>> <div><div class="item"></div><a class="item"></a></div> ❌
    >>> <div><div class="item"></div><a class="widget"></a></div> ✔️
    >>> <div><div class="item"></div></div> ✔️
    >>> <div><div class="widget"></div></div> ❌

    Notes
    -----
    For more information about standard :only-of-type pseudo-class, visit:

    https://developer.mozilla.org/en-US/docs/Web/CSS/:only-of-type
    """

    def __init__(self, selector: SoupSelector) -> None:
        """
        Initializes `OnlyOfSelector` instance.

        Parameters
        ----------
        selector : SoupSelector
            Any `SoupSelector` instance used to match elements.

        Raises
        ------
        NotSoupSelectorException
            If selector is not an instance of `SoupSelector`.
        """
        self.selector = check_selector(selector)

    def find_all(self, tag: Tag, recursive: bool = True, limit=None) -> list[Tag]:
        tag_iterator = (
            TagIterator(tag, recursive=recursive, include_self=True)
            if recursive
            else iter([tag])
        )

        matching = [
            self.selector.find_all(tag=tag_, recursive=False) for tag_ in tag_iterator
        ]
        matches = [elements[0] for elements in matching if len(elements) == 1]

        # keep order of tags and limit
        results = TagResultSet(
            list(TagIterator(tag, recursive=recursive))
        ) & TagResultSet(matches)

        return results.fetch(limit)

    def __eq__(self, other):
        if not isinstance(other, OnlyOfSelector):
            return False

        return self.selector == other.selector

    def __repr__(self):
        cls = self.__class__.__name__
        return f"{cls}(selector={self.selector})"



"""
Module with implementation of templates that complement generator components
providing more customization options.

Classes
-------
- `EmptyTemplate` - Template for generating an empty string.
- `ConstantTemplate` - Template for generating a constant string.
- `ChoiceTemplate` - Template for randomly selecting from a list of choices.
- `RandomTemplate` - Template for generating a random string of ASCII letters and digits.
"""

from .templates import ChoiceTemplate, ConstantTemplate, EmptyTemplate, RandomTemplate

__all__ = [
    "ChoiceTemplate",
    "ConstantTemplate",
    "EmptyTemplate",
    "RandomTemplate",
]



"""Module for custom `soupsavvy` exceptions."""


class SoupsavvyException(Exception):
    """Custom Base Exception for `soupsavvy` specific Exceptions."""


class NotTagSearcherException(SoupsavvyException, TypeError):
    """
    Exception to be raised when function excepted `TagSearcher` as input
    and got argument of the different, invalid type.
    """


#! SELECTORS


class SoupSelectorException(SoupsavvyException):
    """Custom Base Exception for `SoupSelector` specific Exceptions."""


class TagNotFoundException(SoupSelectorException):
    """
    Exception to be raised when `SoupSelector` element was not found in the markup.
    Raised by `SoupSelector` find method when strict parameter is set to True
    and target element was not found in provided `bs4.Tag` object.
    """


class NavigableStringException(SoupSelectorException):
    """
    Exception to be raised when `NavigableString` was returned in `_find` method.
    When `bs4.Tag` find method was used with only 'string' parameter and any tag
    matched, `NavigableString` is returned, which does not have information about the tag.

    Example
    -------
    >>> nav_string = NavigableString("Hello World")
    ... BeautifulSoup(nav_string)
    <html><body><p>Hello World</p></body></html>

    `BeautifulSoup` constructor always wraps NavigableString in <p> element.
    It is rather an unexpected behavior and should rather be avoided in `SoupSelector`
    find method, thus this exception is raised.
    """


class NotSoupSelectorException(SoupSelectorException, TypeError):
    """
    Exception to be raised when function excepted `SoupSelector` as input
    and got argument of the different, invalid type.
    """


class InvalidCSSSelector(SoupsavvyException):
    """
    Raised when the provided CSS selector is invalid.
    Usually raised when invalid formula was passed into 'nth' selector.

    Example
    -------
    >>> NthChild("2x + 1")
    InvalidCSSSelector
    """


class InvalidXPathSelector(SoupsavvyException):
    """
    Raised when the parsing provided XPath selector in `XPathSelector`
    failed due to invalid syntax.

    Example
    -------
    >>> XPathSelector("2x + 1")
    InvalidXPathSelector
    """


#! GENERATORS


class HTMLGeneratorException(SoupsavvyException):
    """Base exception for the HTML generator module."""


class NotUniqueAttributesException(HTMLGeneratorException):
    """
    Exception raised by `TagGenerator` when input attributes have non-unique names.
    HTML tags cannot have multiple attributes with the same name.

    Example
    ------
    >>> TagGenerator(name="a", attrs=["href", ("href", "/endpoint")]
    NotUniqueAttributesException
    """


class VoidTagWithChildrenException(HTMLGeneratorException):
    """
    Exception raised by `TagGenerator` when input name of the tag is a void html tag
    like <img> or <br> and children are specified, which is not an allowed operation.

    Example
    ------
    >>> TagGenerator(name="img", children=["div", "span"])
    VoidTagWithChildrenException
    """


class EmptyNameException(HTMLGeneratorException):
    """
    Exception raised by `TagGenerator` when input name of the tag is empty.
    Exception raised by `AttributeGenerator` when input name of the attribute is empty.

    Empty string is not acceptable in these contexts.

    Example
    ------
    >>> TagGenerator(name="")
    EmptyNamException

    Example
    ------
    >>> AttributeGenerator(name="")
    EmptyNamException
    """


class InvalidTemplateException(HTMLGeneratorException):
    """
    Exception raised by `AttributeGenerator` when invalid template is passed as value
    or by `TagGenerator` when invalid template is passed as text.
    Allowed types are str, `BaseTemplate` and None. Passing other types is not acceptable.

    Example
    ------
    >>> AttributeGenerator(name="class", value=123)
    InvalidTemplateException

    Example
    ------
    >>> TagGenerator(name="div", text=["hello", "world"])
    InvalidTemplateException
    """


class AttributeParsingError(HTMLGeneratorException):
    """
    Exception raised by `TagGenerator` when data passed into attributes iterable
    contains element that couldn't be parsed into `AttributeGenerator` object.
    Relevant only to elements in attributes iterable that
    are not already `AttributeGenerator` objects.

    Example
    ------
    >>> TagGenerator(name="div", attrs=[123]) # not a string
    AttributeParsingError

    Example
    ------
    >>> TagGenerator(name="div", attrs=[("", "test")]) # empty name string
    AttributeParsingError

    Example
    ------
    >>> TagGenerator(name="div", attrs=[("class", ["hello", "soupsavvy"]]) # invalid value template
    AttributeParsingError
    """


#! OPERATIONS


class OperationException(SoupsavvyException):
    """Base exception for the `BaseOperation` related errors"""


class FailedOperationExecution(SoupsavvyException):
    """
    Exception raised by `BaseOperation` when operation execution failed and raised
    any exception.
    """


class NotOperationException(OperationException, TypeError):
    """
    Exception to be raised when function excepted `BaseOperation` as input
    and got argument of the different, invalid type.
    """


class BreakOperationException(OperationException):
    """
    Exception raised by breaking operations such `Break`, to stop the execution
    of the pipeline. Special exception that accepts result as first argument.
    """

    def __init__(self, result) -> None:
        self.result = result


#! MODELS


class BaseModelException(SoupsavvyException):
    """Base exception related to models package and `BaseModel` class."""


class UnknownModelFieldException(BaseModelException):
    """
    Exception raised by `BaseModel` when model passing unknown field name
    to model constructor.
    """


class ModelNotFoundException(BaseModelException):
    """
    Exception raised by `BaseModel` when model scope was not found in provided tag
    and strict parameter was set to True.
    """


class FieldExtractionException(BaseModelException):
    """
    Exception raised by `BaseModel` when field extraction failed and related exception
    was raised by any selector step.
    """


class ScopeNotDefinedException(BaseModelException):
    """
    Exception raised by `BaseModel` when scope was not found in defined model class.
    `__scope__` class attribute is mandatory in `BaseModel` derived classes.
    """


class RequiredConstraintException(BaseModelException):
    """
    Exception raised by `Required` field wrapper when required field was not found
    in provided tag.
    """


class FieldsNotDefinedException(BaseModelException):
    """
    Exception raised by `BaseModel` when fields were not found in defined model class.
    At least one field should be defined in class to be a valid model.
    """


class MissingFieldsException(BaseModelException):
    """
    Exception raised by `BaseModel` when parameters passed into the constructor
    do not contain all required fields defined in the model class.
    """


class FrozenModelException(BaseModelException):
    """
    Exception raised by `BaseModel` when trying to modify frozen model instance.
    Frozen model instances are read-only and cannot be modified.
    """



"""
Module for operators, which are alternative way to create composite selectors
by combining multiple of them in particular way.

Contains
--------
- `is_` - function to create `SelectorList` from multiple selectors
- `where` - alias for `is_`, no difference in behavior in scraping context
- `or_` - alias for `is_`, no difference in behavior in scraping context
- `not_` - function to create `NotSelector` from multiple selectors
- `and_` - function to create `AndSelector` from multiple selectors
- `has` - function to create `HasSelector` from multiple selectors
- `xor` - function to create `XORSelector` from multiple selectors

These utils do not brings extra functionality into the package, but they can be
used for convenience and readability in some cases.
"""

from soupsavvy.base import SoupSelector
from soupsavvy.selectors.logical import (
    AndSelector,
    NotSelector,
    SelectorList,
    XORSelector,
)
from soupsavvy.selectors.relative import HasSelector


def is_(
    selector1: SoupSelector,
    selector2: SoupSelector,
    /,
    *selectors: SoupSelector,
) -> SelectorList:
    """
    Returns an union of multiple soup selectors.

    Parameters
    ----------
    selectors: SoupSelector
         At least two `SoupSelector` objects to match accepted as positional arguments.

    Returns
    -------
    SelectorList
        `SelectorList` object that represents union of provided selectors.

    Example
    -------
    >>> is_(TypeSelector("h1"), TypeSelector("h2"))
    ... where(TypeSelector("h1"), TypeSelector("h2"))

    This is CSS counterpart of `selector list`, `:is()` or `:where` pseudo-classes.

    Example
    -------
    >>> h1, h1
    >>> :is(h1, h2)
    >>> :where(h1, h2)

    Aliases
    -------
    - `where` - to imitate css `:where` pseudo-class
    - `or_` - to imitate logical operator `or`

    Notes
    -----
    For more information on css counterparts, see:

    - https://developer.mozilla.org/en-US/docs/Web/CSS/:is
    - https://developer.mozilla.org/en-US/docs/Web/CSS/:where
    - https://developer.mozilla.org/en-US/docs/Web/CSS/Selector_list
    """
    return SelectorList(selector1, selector2, *selectors)


where = is_
or_ = is_


def not_(selector: SoupSelector, /, *selectors: SoupSelector) -> NotSelector:
    """
    Returns the negation of multiple soup selectors.

    Parameters
    ----------
    selectors: SoupSelector
        `SoupSelector` object(s) to negate match accepted as positional arguments.

    Returns
    -------
    NotSelector
        `NotSelector` object that represents negation of provided selector(s).

    Example
    -------
    >>> not_(TypeSelector("div"), TypeSelector("a"))


    This is the counterpart of CSS `:not()` negation pseudo-class,
    that represents elements that do not match a list of selectors.

    Example
    -------
    >>> :not(div, a)

    Notes
    -----
    For more information on `:not()` pseudo-class, see:

    - https://developer.mozilla.org/en-US/docs/Web/CSS/:not
    """
    return NotSelector(selector, *selectors)


def and_(
    selector1: SoupSelector,
    selector2: SoupSelector,
    /,
    *selectors: SoupSelector,
) -> AndSelector:
    """
    Returns an intersection of multiple soup selectors.
    Element must match all of them to be selected.

    Parameters
    ----------
    selectors: SoupSelector
        At least two `SoupSelector` objects to match accepted as positional arguments.

    Returns
    -------
    AndSelector
        `AndSelector` object that represents intersection of provided selectors.

    Example
    -------
    >>> and_(TypeSelector("div"), ClassSelector("widget"))

    This is the counterpart of CSS compound selector (selector concatenation):

    Example
    -------
    >>> div.widget

    Notes
    -----
    For more information on compound selectors, see:

    - https://developer.mozilla.org/en-US/docs/Web/CSS/CSS_selectors/Selector_structure#compound_selector
    """
    return AndSelector(selector1, selector2, *selectors)


def has(
    selector: SoupSelector,
    /,
    *selectors: SoupSelector,
) -> HasSelector:
    """
    Returns `HasSelector`, whose behavior imitate css `:has()` pseudo-class,
    that represents elements if any of the relative selectors that are passed as an argument
    match at least one element when anchored against this element.

    Parameters
    ----------
    selectors: SoupSelector
        `SoupSelector` object(s) to match accepted as positional arguments.

    Returns
    -------
    HasSelector
        `HasSelector` object that represents `:has()` pseudo-class behavior.

    Example
    -------
    >>> has_(TypeSelector("div"), TypeSelector("a"))

    This is the counterpart of CSS `:has()` pseudo-class:

    Example
    -------
    >>> :has(div, a)

    Notes
    -----
    For more information on `:has()` pseudo-class, see:

    - https://developer.mozilla.org/en-US/docs/Web/CSS/:has
    """
    return HasSelector(selector, *selectors)


def xor(
    selector1: SoupSelector,
    selector2: SoupSelector,
    /,
    *selectors: SoupSelector,
) -> XORSelector:
    """
    Returns an symmetric difference of multiple soup selectors.
    Element must match exactly one of them to be selected.

    Parameters
    ----------
    selectors: SoupSelector
        At least two `SoupSelector` objects to match accepted as positional arguments.

    Returns
    -------
    XORSelector
       `XORSelector` object that represents symmetric difference of provided selectors.

    Example
    -------
    >>> xor(TypeSelector("div"), ClassSelector("widget"))

    This is a shortcut for defining `XOR` operation between two selectors like this:

    Example
    -------
    >>> selector1 = TypeSelector("div")
    ... selector2 = ClassSelector("widget")
    ... xor = (selector1 & (~selector2)) | ((~selector1) & selector2)
    """
    return XORSelector(selector1, selector2, *selectors)
